import coiled

SUCCESS_MESSAGE = """
[bold]Setup complete 🎉[/bold]

What's next?

  Run a command line application in the cloud with:

    $ [bold]coiled run echo 'Hello, world'[/bold]

  Or create a Dask cluster with:

    $ ipython

    [bold]import coiled
    cluster = coiled.Cluster(
        n_workers=10,
    )
    client = cluster.get_client()[/bold]

  For more examples see [link]https://docs.coiled.io/user_guide/examples/index.html[/link]
""".strip()


def setup_failure(reason: str, backend: str):
    coiled.add_interaction("setup-failure", error_message=reason, success=False, backend=backend)



import asyncio
import logging
import platform
import typing
from logging import getLogger
from typing import Dict, List, Optional, Set, Tuple, cast

from rich.progress import Progress
from typing_extensions import Literal

from coiled.context import track_context
from coiled.scan import scan_prefix
from coiled.software_utils import (
    check_pip_happy,
    create_wheels_for_local_python,
    create_wheels_for_packages,
    partition_ignored_packages,
    partition_local_packages,
    partition_local_python_code_packages,
)
from coiled.types import ArchitectureTypesEnum, PackageInfo, PackageLevelEnum, ResolvedPackageInfo, parse_conda_channel
from coiled.v2.core import CloudV2
from coiled.v2.widgets.util import simple_progress

PYTHON_VERSION = platform.python_version_tuple()
ANY_AVAILABLE = "ANY-AVAILABLE"


logger = getLogger("coiled.package_sync")


async def default_python() -> PackageInfo:
    python_version = platform.python_version()
    return {
        "name": "python",
        "path": None,
        "source": "conda",
        "channel_url": ANY_AVAILABLE,
        "channel": ANY_AVAILABLE,
        "subdir": "linux-64",
        "conda_name": "python",
        "version": python_version,
        "wheel_target": None,
    }


@track_context
async def approximate_packages(
    cloud: CloudV2,
    packages: List[PackageInfo],
    priorities: Dict[Tuple[str, Literal["conda", "pip"]], PackageLevelEnum],
    progress: Optional[Progress] = None,
    strict: bool = False,
    architecture: ArchitectureTypesEnum = ArchitectureTypesEnum.X86_64,
    pip_check_errors: Optional[Dict[str, List[str]]] = None,
    gpu_enabled: bool = False,
) -> typing.List[ResolvedPackageInfo]:
    user_conda_installed_python = next((p for p in packages if p["name"] == "python"), None)
    user_conda_installed_pip = next(
        (i for i, p in enumerate(packages) if p["name"] == "pip" and p["source"] == "conda"),
        None,
    )
    if not user_conda_installed_pip:
        # This means pip was installed by pip, or the system
        # package manager
        # Insert a conda version of pip to be installed first, it will
        # then be used to install the users version of pip
        pip = next(
            (p for p in packages if p["name"] == "pip" and p["source"] == "pip"),
            None,
        )
        if not pip:
            # insert a modern version and hope it does not introduce conflicts
            packages.append({
                "name": "pip",
                "path": None,
                "source": "conda",
                "channel_url": "https://conda.anaconda.org/conda-forge/",
                "channel": "conda-forge",
                "subdir": "noarch",
                "conda_name": "pip",
                "version": "22.3.1",
                "wheel_target": None,
            })
        else:
            # insert the users pip version and hope it exists on conda-forge
            packages.append({
                "name": "pip",
                "path": None,
                "source": "conda",
                "channel_url": "https://conda.anaconda.org/conda-forge/",
                "channel": "conda-forge",
                "subdir": "noarch",
                "conda_name": "pip",
                "version": pip["version"],
                "wheel_target": None,
            })
    coiled_selected_python = None
    if not user_conda_installed_python:
        # insert a special python package
        # that the backend will pick a channel for
        coiled_selected_python = await default_python()
        packages.append(coiled_selected_python)
    packages, _ = partition_ignored_packages(packages, priorities=priorities)
    packages, local_python_code = partition_local_python_code_packages(packages)
    packages, local_python_wheel_packages = partition_local_packages(packages)
    with simple_progress("Validating environment", progress=progress):
        results = await cloud._approximate_packages(
            packages=[
                {
                    "name": pkg["name"],
                    "priority_override": (
                        PackageLevelEnum.CRITICAL
                        if (
                            strict
                            or (
                                pkg["wheel_target"]
                                # Ignore should override wheel_target (see #2640)
                                and not priorities.get((pkg["name"], pkg["source"])) == PackageLevelEnum.IGNORE
                            )
                        )
                        else priorities.get((
                            (cast(str, pkg["conda_name"]) if pkg["source"] == "conda" else pkg["name"]),
                            pkg["source"],
                        ))
                    ),
                    "python_major_version": PYTHON_VERSION[0],
                    "python_minor_version": PYTHON_VERSION[1],
                    "python_patch_version": PYTHON_VERSION[2],
                    "source": pkg["source"],
                    "channel_url": pkg["channel_url"],
                    "channel": pkg["channel"],
                    "subdir": pkg["subdir"],
                    "conda_name": pkg["conda_name"],
                    "version": pkg["version"],
                    "wheel_target": pkg["wheel_target"],
                }
                # Send all packages to backend to help with debugging
                for pkg in packages + local_python_code + local_python_wheel_packages
            ],
            architecture=architecture,
            pip_check_errors=pip_check_errors,
            gpu_enabled=gpu_enabled,
        )
    finalized_packages: typing.List[ResolvedPackageInfo] = []
    finalized_packages.extend(await create_wheels_for_local_python(local_python_code, progress=progress))
    finalized_packages.extend(await create_wheels_for_packages(local_python_wheel_packages, progress=progress))
    for package_result in results:
        # Ensure channel from package_result is a channel and not a URL for conda packages
        if package_result["conda_name"] and package_result["channel_url"]:
            subdir = f"linux-{architecture.conda_suffix}"
            channel = parse_conda_channel(package_result["name"], package_result["channel_url"], subdir)[0]
        else:
            channel = package_result["channel_url"]
        # Remove channel_url note that endpoint returns for backward compatibility
        if (
            channel
            and package_result["note"]
            and (channel == package_result["note"] or package_result["note"].endswith(f",{channel}"))
        ):
            package_result["note"] = None

        finalized_packages.append({
            "name": package_result["name"],
            "source": "conda" if package_result["conda_name"] else "pip",
            "channel": channel,
            "conda_name": package_result["conda_name"],
            "client_version": package_result["client_version"],
            "specifier": package_result["specifier"] or "",
            "include": package_result["include"],
            "note": package_result["note"],
            "error": package_result["error"],
            "sdist": None,
            "md5": None,
        })

    return finalized_packages


@track_context
async def create_environment_approximation(
    cloud: CloudV2,
    priorities: Dict[Tuple[str, Literal["conda", "pip"]], PackageLevelEnum],
    only: Optional[Set[str]] = None,
    strict: bool = False,
    progress: Optional[Progress] = None,
    architecture: ArchitectureTypesEnum = ArchitectureTypesEnum.X86_64,
    gpu_enabled: bool = False,
) -> typing.List[ResolvedPackageInfo]:
    packages = await scan_prefix(progress=progress)
    pip_check_errors = await check_pip_happy(progress)
    if only:
        packages = [pkg for pkg in packages if pkg["name"] in only]
    result = await approximate_packages(
        cloud=cloud,
        packages=[pkg for pkg in packages],
        priorities=priorities,
        strict=strict,
        progress=progress,
        architecture=architecture,
        pip_check_errors=pip_check_errors,
        gpu_enabled=gpu_enabled,
    )
    return result


if __name__ == "__main__":
    from logging import basicConfig

    basicConfig(level=logging.INFO)

    from rich.console import Console
    from rich.table import Table

    async def run():
        async with CloudV2(asynchronous=True) as cloud:
            return await create_environment_approximation(
                cloud=cloud,
                priorities={
                    ("dask", "conda"): PackageLevelEnum.CRITICAL,
                    ("twisted", "conda"): PackageLevelEnum.IGNORE,
                    ("graphviz", "conda"): PackageLevelEnum.LOOSE,
                    ("icu", "conda"): PackageLevelEnum.LOOSE,
                },
            )

    result = asyncio.run(run())

    table = Table(title="Packages")
    keys = ("name", "source", "include", "client_version", "specifier", "error", "note")

    for key in keys:
        table.add_column(key)

    for pkg in result:
        row_values = [str(pkg.get(key, "")) for key in keys]
        table.add_row(*row_values)
    console = Console()
    console.print(table)
    console.print(table)



import click

from ..utils import CONTEXT_SETTINGS
from .azure_logs import show_azure_logs
from .better_logs import better_logs_cli as better_logs
from .crud import stop
from .list import list_clusters
from .logs import logs
from .ssh import ssh


@click.group(context_settings=CONTEXT_SETTINGS)
def cluster():
    """Commands for managing Coiled clusters"""
    pass


cluster.add_command(stop)
cluster.add_command(ssh)
cluster.add_command(logs, "logs-via-aws-cli")
cluster.add_command(better_logs, "logs")
cluster.add_command(better_logs)
cluster.add_command(list_clusters, "list")
cluster.add_command(show_azure_logs, "azure-logs")



import click
from rich.console import Console
from rich.table import Table

import coiled

from ...utils import get_details_url
from ..utils import CONTEXT_SETTINGS

console = Console()


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option(
    "--workspace", default=None, required=False, help="Coiled workspace (uses default workspace if not specified)."
)
@click.option(
    "--just-mine",
    default=False,
    is_flag=True,
    help="Show only my clusters",
)
@click.option(
    "--max-pages",
    default=1,
    type=int,
    help="Maximum number of pages to show (where each page is 100 clusters)",
)
def list_clusters(workspace: str, just_mine: bool, max_pages: int):
    """List the Coiled clusters in a workspace"""
    with coiled.Cloud(workspace=workspace) as cloud:
        clusters = cloud.list_clusters(
            workspace=workspace,
            just_mine=just_mine,
            max_pages=max_pages,
        )
        table = Table(title="Clusters")
        table.add_column("Name", style="cyan")
        table.add_column("Updated", style="magenta")
        table.add_column("Link", style="magenta")
        table.add_column("Status", style="green")
        for cluster_details in clusters:
            table.add_row(
                cluster_details["name"],
                cluster_details["updated"],
                get_details_url(cloud.server, workspace or cloud.default_workspace, cluster_details["id"]),
                cluster_details["current_state"].get("state") or "unknown",
            )
        console.print(table)



from typing import Optional

import click

import coiled

from ..utils import CONTEXT_SETTINGS
from .utils import find_cluster


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument("cluster", default="", required=False)
@click.option(
    "--account",
    "--workspace",
    default=None,
    help="Coiled workspace (uses default workspace if not specified)."
    " Note: --account is deprecated, please use --workspace instead.",
)
def stop(
    cluster: str,
    account: Optional[str],
):
    with coiled.Cloud(account=account) as cloud:
        cluster_info = find_cluster(cloud, cluster)
        print(
            f"Requesting stop for {cluster_info['name']} ({cluster_info['id']}), "
            f"current state is {cluster_info['current_state']['state']}"
        )
        cluster_id = cluster_info["id"]
        cloud.delete_cluster(cluster_id, account, reason="User requested stop via CLI")



from __future__ import annotations

import asyncio
import logging
import signal
import subprocess
from pathlib import Path
from typing import List, Optional, Sequence, Union
from webbrowser import open as open_webbrowser

import click
import importlib_metadata
from packaging.version import Version
from rich import print
from rich.control import Control
from rich.prompt import Confirm
from rich.segment import ControlType
from urllib3.util import parse_url

import coiled
from coiled.cli.cluster.ssh import check_ssh_agent
from coiled.compatibility import DISTRIBUTED_VERSION
from coiled.errors import DoesNotExist
from coiled.software_utils import make_coiled_local_name
from coiled.utils import dict_from_key_val_list, error_info_for_tracking

from ...utils import unset_single_thread_defaults
from ...v2.widgets.rich import LightRichClusterWidget
from ..sync import (
    MUTAGEN_NAME_FORMAT,
    can_use_mutagen,
    connect_mutagen_sync,
    get_mutagen_path,
    mutagen_session_exists,
    start_sync,
    stop_sync,
)
from ..utils import CONTEXT_SETTINGS

# Path on VM to sync to.
# We use `/scratch` for now because it's already bind-mounted into docker.
MIN_DISTRIBUTED_VERSION = Version("2022.8.1")


class NotebookError(Exception):
    pass


def check_distributed_version() -> bool:
    if DISTRIBUTED_VERSION < MIN_DISTRIBUTED_VERSION:
        print(
            "[bold red]"
            f"distributed>{MIN_DISTRIBUTED_VERSION} is required to launch notebooks. "
            f"You have {DISTRIBUTED_VERSION}."
            "[/]"
        )
        return False
    return True


def check_jupyter() -> bool:
    try:
        importlib_metadata.distribution("jupyterlab")
    except ModuleNotFoundError:
        print("[bold red]Jupyterlab must be installed locally to launch notebooks.[/]")
        return False

    try:
        importlib_metadata.distribution("jupyter_server_proxy")
    except ModuleNotFoundError:
        print(
            "[bold red]jupyter-server-proxy is not installed, "
            "without this you won't be able to access Dask dashboard for local clusters created on notebook server.[/]"
        )

    return True


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option(
    "--name",
    default=None,
    help="Cluster name. If not given, defaults to a hash based on current working directory.",
)
@click.option(
    "--account",
    "--workspace",
    default=None,
    help="Coiled workspace (uses default workspace if not specified)."
    " Note: --account is deprecated, please use --workspace instead.",
)
@click.option(
    "--sync",
    default=False,
    is_flag=True,
    help="Sync the working directory with the filesystem on the notebook. Requires mutagen.",
)
@click.option(
    "--allow-sync-from",
    default="me",
    type=str,
    help=(
        "IP address or CIDR from which connections to port 22 (SSH) are open; "
        "can also be specified as 'everyone' (0.0.0.0/0) or 'me' (automatically determines public IP detected "
        "for your local client). Note that `--sync` relies on SSH connection."
    ),
)
@click.option(
    "--software",
    default=None,
    help=(
        "Software environment name to use. If neither software nor container is specified, "
        "all the currently-installed Python packages are replicated on the VM using package sync."
    ),
)
@click.option(
    "--container",
    default=None,
    help=(
        "Container image to use. If neither software nor container is specified, "
        "all the currently-installed Python packages are replicated on the VM using package sync."
    ),
)
@click.option(
    "--vm-type",
    default=[],
    multiple=True,
    help="VM type to use. Specify multiple times to provide multiple options.",
)
@click.option("--cpu", default=None, type=int, help="Number of CPUs requested for this notebook.")
@click.option(
    "--memory",
    default=None,
    help="Amount of memory for this notebook, Coiled will use a +/-10% buffer from the memory you specify.",
)
@click.option(
    "--gpu",
    default=False,
    is_flag=True,
    help="Use GPU notebook server.",
)
@click.option(
    "--disk-size",
    default=None,
    help="Use larger-than-default disk on VM, specified in GiB.",
)
@click.option(
    "--region",
    default=None,
    help="The cloud provider region in which to run the notebook.",
)
@click.option(
    "--open",
    default=True,
    is_flag=True,
    help="Whether to open the notebook in the default browser once it's launched.",
)
@click.option(
    "--block/--no-block",
    default=True,
    is_flag=True,
    help="Whether to block while the notebook is running.",
)
@click.option(
    "--include-vcs",
    default=False,
    is_flag=True,
    help="Include version control files when syncing (e.g., `.git`).",
)
@click.option(
    "--idle-timeout",
    default="4 hours",
    help="Period of idleness after which to automatically shut down the notebook, "
    "e.g. '20 minutes' or '1 hour' or '10 days' (default is '4 hours'). "
    "The notebook is considered active (not idle) if a browser is connected, "
    "even if no code is running.",
)
@click.option(
    "--allow-ingress-from",
    default=None,
    help="Control the CIDR from which cluster firewall allows ingress to scheduler; "
    "by default this is open to any source address (0.0.0.0/0). You can specify "
    "CIDR, or 'me' for just your IP address.",
)
@click.option(
    "--env",
    "-e",
    default=[],
    multiple=True,
    help=(
        "Environment variables securely transmitted to run command environment. "
        "Format is ``KEY=val``, multiple vars can be set with separate ``--env`` for each."
    ),
)
@click.option(
    "--tag",
    "-t",
    default=[],
    multiple=True,
    help="Tags. Format is ``KEY=val``, multiple vars can be set with separate ``--tag`` for each.",
)
@click.option(
    "--sync-ignore",
    default=None,
    multiple=True,
    help="Paths to not sync when using ``--sync``.",
)
@click.option(
    "--mount-bucket",
    default=None,
    multiple=True,
    help="S3 or GCS bucket(s) to mount as volumes.",
)
def start_notebook(
    name: Optional[str],
    account: Optional[str],
    sync: bool,
    allow_sync_from: str,
    software: Optional[str],
    container: Optional[str],
    vm_type: Sequence[str],
    cpu: Union[int, None],
    memory: Union[str, None],
    gpu: bool,
    disk_size: Optional[int],
    region: Optional[str],
    open: bool,
    block: bool,
    include_vcs: bool,
    idle_timeout: str,
    allow_ingress_from: Optional[str],
    env: Optional[List[str]],
    tag: Optional[List[str]],
    sync_ignore: Optional[List[str]],
    mount_bucket: Optional[List[str]],
):
    """
    Launch or re-open a notebook session, with optional file syncing.

    If a notebook session with the same ``name`` already exists, it's not re-created.
    If file sync was initially not enabled, running ``coiled notebook start --sync``
    will begin file sync without re-launching the notebook.
    """
    _start_notebook(
        name=name,
        workspace=account,
        sync=sync,
        allow_sync_from=allow_sync_from,
        software=software,
        container=container,
        vm_type=vm_type,
        cpu=cpu,
        memory=memory,
        gpu=gpu,
        disk_size=disk_size,
        region=region,
        open=open,
        block=block,
        include_vcs=include_vcs,
        idle_timeout=idle_timeout,
        allow_ingress_from=allow_ingress_from,
        env=env,
        tag=tag,
        sync_ignore=sync_ignore,
        mount_bucket=mount_bucket,
    )


def _start_notebook(
    *,
    name: Optional[str] = None,
    workspace: Optional[str] = None,
    sync: bool = False,
    allow_sync_from: str = "me",
    software: Optional[str] = None,
    container: Optional[str] = None,
    vm_type: Sequence[str] = [],
    cpu: Union[int, None] = None,
    memory: Union[str, None] = None,
    gpu: bool = False,
    disk_size: Optional[int] = None,
    region: Optional[str] = None,
    open: bool = False,
    block: bool = True,
    include_vcs: bool = False,
    idle_timeout: str = "4 hours",
    debug: bool = False,
    allow_ingress_from: Optional[str] = None,
    env: Optional[List[str]] = None,
    tag: Optional[List[str]] = None,
    sync_ignore: Optional[List[str]] = None,
    mount_bucket: Optional[List[str]] = None,
) -> Optional[coiled.Cluster]:
    """
    Launch or re-open a notebook session, with optional file syncing.

    If a notebook session with the same ``name`` already exists, it's not re-created.
    If file sync was initially not enabled, running ``coiled notebook start --sync``
    will begin file sync without re-launching the notebook.
    """
    info = {"vm_type": vm_type, "sync": sync, "block": block, "include_vcs": include_vcs}
    success = True
    exception = None

    runtime_env_dict = dict_from_key_val_list(env)
    tags = dict_from_key_val_list(tag)

    try:
        # when using package sync, check that local env has jupyter and recent distributed
        if not software and not container:
            if not check_distributed_version():
                raise NotebookError("distributed version check")
            if not check_jupyter():
                raise NotebookError("jupyter missing")

        if sync and not can_use_mutagen():
            return

        if sync and not check_ssh_agent():
            print("Sync may be unable to start but we'll try anyway...")
            info["problem"] = "check_ssh_agent failed"

        initial_env = unset_single_thread_defaults()
        if container and "rapidsai" in container:
            initial_env["DISABLE_JUPYTER"] = "true"  # needed for "stable" RAPIDS image

        name = name or f"notebook-{coiled.utils.short_random_string()}"
        with LightRichClusterWidget(
            workspace=workspace,
            title=f"Notebook [bold]{name}[/bold]...",
            extra_link_title="Jupyter",
            extra_link="...",
            width=84,
        ) as widget, coiled.Cloud(workspace=workspace) as cloud:
            workspace = workspace or cloud.default_workspace
            info["workspace"] = workspace
            widget.update(
                server=cloud.server,
                cluster_details=None,
                logs=None,
                workspace=workspace,
                extra_link="...",
            )
            cluster = coiled.Cluster(
                name=name,
                workspace=workspace,
                cloud=cloud,
                n_workers=0,
                software=software,
                container=container,
                jupyter=True,
                scheduler_options={"idle_timeout": idle_timeout},
                scheduler_vm_types=list(vm_type) if vm_type else None,
                worker_vm_types=list(vm_type) if vm_type else None,
                scheduler_cpu=cpu,
                scheduler_memory=memory,
                scheduler_disk_size=disk_size,
                allow_ssh_from=allow_sync_from,
                environ=initial_env,
                scheduler_gpu=gpu,
                region=region,
                tags={**tags, "coiled-cluster-type": "notebook"},
                custom_widget=widget,
                package_sync_ignore=[make_coiled_local_name(Path.cwd().resolve().name)],
                allow_ingress_from=allow_ingress_from,
                mount_bucket=mount_bucket,
            )
            info["cluster_id"] = cluster.cluster_id

            url = cluster.jupyter_link
            cluster_id = cluster.cluster_id
            assert cluster_id is not None

            # by default, jupyter on the scheduler gives us client to that very scheduler
            # clear ENV var so default `Client()` on notebook gives us a new local cluster
            cluster.unset_env_vars(["DASK_SCHEDULER_ADDRESS"])

            # some env vars are set in dask config early in the dask loading process
            # we set these to `""` since that's what we can do in the relevant dask config
            # but some libraries don't accept empty string, so we'll unset them here.
            # this won't affect things loaded by dask, but should affect user-code.
            cluster.unset_env_vars(unset_single_thread_defaults().keys())

            if runtime_env_dict:
                cluster.send_private_envs(runtime_env_dict)

            if sync:
                url = parse_url(url)._replace(path="/jupyter/lab/tree/synced").url

            widget.update(
                cluster_details=None,
                logs=None,
                extra_link=url,
            )

            if sync:
                connected, error_message = connect_mutagen_sync(
                    cloud,
                    cluster_id,
                    widget.live.console,
                    include_vcs=include_vcs,
                    debug=debug,
                    ignores=sync_ignore,
                )
                if not connected:
                    widget.stop()
                    print(
                        "[red]Unable to start sync, see error shown above for more details.[/red]\n"
                        "If sync is not required, you can start notebook without [green]--sync[/green]."
                    )

                    cluster._queue_cluster_event(
                        "notebook", f"error attempting to connect sync:\n{error_message}", level=logging.ERROR
                    )

                    _stop_notebook(name=name, workspace=workspace, cluster_id=cluster_id)

                    # close the cluster object so we comms close cleanly (and don't print error)
                    cluster.close(reason="The notebook sync failed to start")
                    raise NotebookError(f"unable to start sync: {error_message}")

            if open:
                open_webbrowser(url, new=2)

            if block:
                widget.update(
                    cluster_details=None,
                    logs=None,
                    trailer="Use Control-C to stop this notebook server",
                )

                def signal_handler_noop(_, frame):
                    # Ignore the input signal
                    return

                async def update_widget():
                    while True:
                        cluster_details = await cloud._get_cluster_details(cluster_id=cluster_id, workspace=workspace)
                        widget.update(cluster_details, logs=None)
                        # don't make user hit control-c if the cluster been stopped another way
                        if cluster_details["scheduler"]["current_state"]["state"] in ("error", "stopped"):
                            return False
                        await asyncio.sleep(1.0)

                while True:
                    try:
                        if not cluster.sync(update_widget):
                            widget.stop()
                            break
                    except KeyboardInterrupt:
                        widget.stop()
                        try:
                            exit = Confirm.ask("Are you sure you want to stop this notebook server?", default=True)
                        except KeyboardInterrupt:
                            exit = True
                        if exit:
                            # Register noop handler since we're shutting down and
                            # want to make sure the notebook is shutdown even when
                            # hammering ctrl-C
                            signal.signal(signal.SIGINT, signal_handler_noop)
                            break
                        else:
                            # write over the prompt
                            print(Control.move(0, -1), Control((ControlType.ERASE_IN_LINE, 2)))
                            print("[green]Continuing with this notebook server... [/]")

                            # move cursor to bottom of widget so widget won't shift
                            print(Control.move(0, -2))
                            widget.start()

                _stop_notebook(name=name, workspace=workspace, cluster_id=cluster_id)

                # close the cluster object so we comms close cleanly (and don't print error)
                cluster.close()
            else:
                stop_command = "coiled notebook stop"
                if workspace:
                    stop_command = f"{stop_command} --workspace {workspace}"
                stop_command = f"{stop_command} {name}"

                print(f"To stop this notebook server: [green]{stop_command}[/]")
        return cluster
    except NotebookError as e:
        success = False
        exception = e
        # don't raise, we've already printed appropriate message about the issue
    except Exception as e:
        success = False
        exception = e
        raise e
    finally:
        coiled.add_interaction(
            "coiled-notebook",
            success=success,
            **info,
            **error_info_for_tracking(exception),
        )


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument("name")
@click.option(
    "--account",
    "--workspace",
    default=None,
    help="Coiled workspace (uses default workspace if not specified)."
    " Note: --account is deprecated, please use --workspace instead.",
)
def stop_notebook(name: str, account: Optional[str]):
    """
    Shut down a notebook session
    """
    _stop_notebook(name=name, workspace=account)


def _stop_notebook(name: str, workspace: Optional[str], cluster_id: Optional[int] = None):
    with coiled.Cloud(workspace=workspace) as cloud:
        notebook_running = True
        try:
            cluster_id = cloud.get_cluster_by_name(name)
        except DoesNotExist:
            if cluster_id:
                print(f"[bold red]Notebook {name!r} is already stopped[/]")
            else:
                print(f"[bold red]Notebook {name!r} does not exist[/]")
            notebook_running = False

        stop_sync(cloud, cluster_id, f"Stopping sync with notebook {name!r} ({cluster_id})")

        if cluster_id and notebook_running:
            print(f"Stopping notebook {name!r} ({cluster_id})...")
            cloud.delete_cluster(cluster_id, workspace, reason="User requested notebook stop via CLI")


@click.command(context_settings=CONTEXT_SETTINGS, hidden=True)
@click.argument("name")
@click.option(
    "--include-vcs",
    default=False,
    is_flag=True,
    help="Include version control files when syncing (e.g., `.git`).",
)
@click.option(
    "--debug",
    default=False,
    is_flag=True,
)
def start_standalone_sync(name: str, include_vcs: bool, debug: bool):
    error = None
    with coiled.Cloud() as cloud:
        try:
            cluster_id = cloud.get_cluster_by_name(name)
            error = start_sync(cloud=cloud, cluster_id=cluster_id, include_vcs=include_vcs, debug=debug)
        except DoesNotExist:
            print(f"[bold red]Cluster {name!r} does not exist[/]")
            return

    coiled.add_interaction(
        "coiled-notebook-standalone-sync",
        success=not error,
        include_vcs=include_vcs,
    )


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument("name")
def monitor_sync(name: str):
    """Monitor file sync status for a notebook session."""
    mutagen_path = get_mutagen_path()
    if not mutagen_path:
        return

    with coiled.Cloud() as cloud:
        try:
            cluster_id = cloud.get_cluster_by_name(name)
        except DoesNotExist:
            print(f"[bold red]Cluster {name!r} does not exist[/]")
            return  # TODO exit 1

    if not mutagen_session_exists(cluster_id):
        print(f"[bold red]No file synchronization session for cluster {name!r} ({cluster_id})[/]")
        return  # TODO exit 1

    subprocess.run([mutagen_path, "sync", "monitor", MUTAGEN_NAME_FORMAT.format(cluster_id=cluster_id)])



import datetime
import json
import shlex
import shutil
import subprocess
import sys
import time
from typing import Optional

import click
import httpx
import rich.prompt
from rich import print
from rich.live import Live
from rich.markup import escape
from rich.panel import Panel
from rich.prompt import Confirm

import coiled

from ...auth import get_local_user
from ..utils import CONTEXT_SETTINGS
from .util import SUCCESS_MESSAGE, setup_failure

try:
    from azure.core.exceptions import ClientAuthenticationError  # type: ignore
    from azure.identity import DefaultAzureCredential  # type: ignore
    from azure.mgmt.resource import ResourceManagementClient  # type: ignore
    from azure.mgmt.subscription import SubscriptionClient  # type: ignore

    AZURE_DEPS = True
except ImportError:
    ClientAuthenticationError = Exception
    AZURE_DEPS = False


REQUIRED_PROVIDERS = ("Microsoft.Network", "Microsoft.Compute", "Microsoft.Storage")
RG_ROLE_NAME = "Coiled Resource Group Role"
LOG_ROLE_NAME = "Coiled Log Access"


def rg_role_def(sub_id):
    return f"""{{
  "Name": "{RG_ROLE_NAME}",
  "IsCustom": true,
  "Description": "Setup and ongoing Coiled permissions required at resource group scope",
  "Actions": [
    "Microsoft.Compute/*/read",
    "Microsoft.Compute/virtualMachines/delete",
    "Microsoft.Compute/virtualMachineScaleSets/*",
    "Microsoft.Network/*/read",
    "Microsoft.Network/applicationSecurityGroups/*",
    "Microsoft.Network/networkSecurityGroups/*",
    "Microsoft.Network/publicIPAddresses/delete",
    "Microsoft.Network/publicIPAddresses/write",
    "Microsoft.Network/virtualNetworks/subnets/join/action",
    "Microsoft.Network/virtualNetworks/subnets/write",
    "Microsoft.Network/virtualNetworks/write",
    "Microsoft.Storage/storageAccounts/managementPolicies/write",
    "Microsoft.Storage/storageAccounts/write"
  ],
  "NotActions": [

  ],
  "AssignableScopes": [
    "/subscriptions/{sub_id}"
  ]
}}""".replace('"', '\\"')


def log_role_def(sub_id):
    return f"""{{
  "Name": "{LOG_ROLE_NAME}",
  "IsCustom": true,
  "Description": "Role needs resource group scope for setup, then storage account scope for on-going",
  "Actions": [
    "Microsoft.Storage/storageAccounts/read",
    "Microsoft.Storage/storageAccounts/listkeys/action"
  ],
  "NotActions": [

  ],
  "AssignableScopes": [
    "/subscriptions/{sub_id}"
  ]
}}""".replace('"', '\\"')


def bash_script(coiled_account, sub_id, rg_name):
    header = f"""#!/bin/bash
COILED_ACCOUNT={coiled_account}
SUBSCRIPTION={sub_id}
RG_NAME={rg_name}"""

    body = """
# you can use any app name, but we'll default to a name that includes coiled account
APP_NAME="coiled-${COILED_ACCOUNT}-app"
RG_ID=/subscriptions/${SUBSCRIPTION}/resourceGroups/${RG_NAME}

# log in if necessary
az ad signed-in-user show 2>&1 1>\\dev\\null || az login

# create app, creds, and service principal
az ad app create --display-name $APP_NAME --query "id" | tr -d '"' > app-id.txt
az ad app credential reset --id $(cat app-id.txt) | tee app-creds.json
az ad sp create --id $(cat app-id.txt) --query "id" | tr -d '"' > sp-id.txt

# create roles
# TODO var substitution for subscription id in json role definitions
az role definition create --role-definition ./coiled-resource-group-role.json
az role definition create --role-definition ./coiled-log-access-role.json

# grant service principle access to the resource group
az role assignment create --role "Coiled Resource Group Role" --scope $RG_ID --assignee $(cat sp-id.txt)
az role assignment create --role "Coiled Log Access" --scope $RG_ID --assignee $(cat sp-id.txt)

# send creds to coiled

APP_PASS=$(jq -r '.password' app-creds.json)
APP_ID=$(jq -r '.appId' app-creds.json)
TENANT=$(jq -r '.tenant' app-creds.json)
SETUP_ENDPOINT=/api/v2/cloud-credentials/${COILED_ACCOUNT}/azure

coiled curl -X POST "${SETUP_ENDPOINT}" --json --data "{\\"credentials\\": {\\"tenant_id\\": \\"${TENANT}\\", \\"client_id\\": \\"${APP_ID}\\", \\"client_secret\\": \\"${APP_PASS}\\"}, \\"subscription_id\\":\\"${SUBSCRIPTION}\\",\\"resource_group_name\\":\\"${RG_NAME}\\"}"
"""  # noqa: E501

    return f"{header}\n{body}"


@click.option(
    "--subscription",
    default=None,
    help=(
        "Azure subscription to use, specified by ID. "
        "(You'll be prompted with options not specified and more than one subscription is found)."
    ),
)
@click.option(
    "--resource-group",
    default=None,
    help=(
        "Azure resource group to use. Note that all permissions are scoped to the resource group."
        "(You'll be prompted with options not specified and more than one resource group is found)."
    ),
)
@click.option("--region", default=None, help="Default region for Coiled to use.")
@click.option(
    "--account",
    "--workspace",
    default=None,
    help=(
        "Coiled workspace (uses default workspace if not specified). "
        "Note: ``--account`` is deprecated, please use ``--workspace`` instead."
    ),
)
@click.option(
    "--iam-user",
    default=None,
    help=(
        "Name of enterprise application/service principal to create. "
        "By default, we'll use ``coiled-<coiled workspace slug>-app``."
    ),
)
@click.option(
    "--save-script",
    is_flag=True,
    default=False,
    hidden=True,
)
@click.option(
    "--ship-token",
    is_flag=True,
    default=False,
    help=(
        "Instead of using a service principal to grant long-term access to your Azure account, this will ship "
        "(and refresh) temporary OAuth tokens so that Coiled can operate in your Azure account as you. "
        "Not recommended for long-term Coiled use, but can be helpful if you don't want to (or can't) create a "
        "service principal for Coiled to use."
    ),
)
@click.command(context_settings=CONTEXT_SETTINGS)
def azure_setup(subscription, resource_group, region, account, iam_user, save_script, ship_token):
    print(
        "Coiled on Azure is currently in [bold]public beta[/bold], "
        "please contact [link]support@coiled.io[/link] if you have any questions or problems."
    )
    local_user = get_local_user()

    try:
        with coiled.Cloud(workspace=account) as cloud:
            coiled_account = account or cloud.default_workspace
    except PermissionError as e:
        print(e)
        return

    coiled.add_interaction(
        action="CliSetupAzure",
        success=True,
        local_user=local_user,
        # use keys that match the cli args
        region=region,
        save_script=save_script,
        ship_token=ship_token,
    )

    has_requirements = True
    has_az_cli = get_cli_path() is not None
    if not AZURE_DEPS:
        print()
        print(
            "[red]Missing requirement[/red] Azure Python libraries are not installed.\n"
            "You can install using [bold]pip[/bold]:\n"
            "  [green]pip install 'coiled\\[azure]'[/green]\n"
            "or [bold]conda[/bold]:\n"
            "  [green]conda install azure-identity azure-mgmt-resource azure-mgmt-subscription -c conda-forge[/green]"
        )
        has_requirements = False

    if not ship_token and not has_az_cli:
        print()
        print("[red]Missing requirement[/red] Unable to find the [green]az[/green] Azure CLI, which this script uses.")
        print(
            "See [link=https://learn.microsoft.com/en-us/cli/azure/install-azure-cli]"
            "https://learn.microsoft.com/en-us/cli/azure/install-azure-cli[/link] "
            "for Azure CLI installation instructions."
        )
        has_requirements = False

    if not has_requirements:
        setup_failure(f"missing requirements:cli={has_az_cli},deps={AZURE_DEPS}", backend="azure")
        return

    creds = DefaultAzureCredential()  # pyright: ignore[reportPossiblyUnboundVariable]
    if not whoami(creds):
        print()
        print("[red]Unable to find local Azure credentials")
        print("We recommend using the Azure CLI to configure your local Azure credentials")
        if not has_az_cli:
            print(
                "See [link=https://learn.microsoft.com/en-us/cli/azure/install-azure-cli]"
                "https://learn.microsoft.com/en-us/cli/azure/install-azure-cli[/link] "
                "for Azure CLI installation instructions."
            )
            print("Once you have the Azure CLI installed, you can run this to configure local credentials:")
        else:
            print("Run this command to configure local credentials:")

        print("  [green]az login")
        setup_failure("no local credentials", backend="azure")
        return

    sub_id = subscription or get_subscription(creds)

    if not sub_id:
        return

    rg_name, rg_location, rg_id = get_rg(creds, sub_id, rg_name=resource_group or "")
    region = region or rg_location

    if not rg_id:
        setup_failure("no resource group selected", backend="azure")
        return

    print(
        f"Coiled account [green]{coiled_account}[/green] will be configured to use "
        f"resource group [green]{rg_name}[/green] "
        f"with [green]{region}[/green] as the default region\n"
    )

    if ship_token:
        enable_providers(creds, sub_id)
        ship_token_creds(
            local_credentials=creds, coiled_account=coiled_account, sub_id=sub_id, rg_name=rg_name, region=region
        )
        return

    if save_script:
        path = "coiled-resource-group-role.json"
        with open(path, "w") as f:
            f.write(rg_role_def(sub_id))
            print(f"Saved [bold]{RG_ROLE_NAME}[/bold] as [green]{path}")
        path = "coiled-log-access-role.json"
        with open(path, "w") as f:
            f.write(log_role_def(sub_id))
            print(f"Saved [bold]{LOG_ROLE_NAME}[/bold] as [green]{path}")
        path = "coiled-setup.sh"
        with open(path, "w") as f:
            f.write(bash_script(coiled_account, sub_id, rg_name))
            print(f"Saved [bold]Coiled setup script[/bold] as [green]{path}")

    else:
        app_name = iam_user or f"coiled-{coiled_account}-app"
        try:
            if not setup_with_service_principal(creds, app_name, sub_id, rg_name, rg_id, coiled_account, region):
                coiled.add_interaction(action="CoiledSetup", success=False)
        except Exception as e:
            error_message = str(e)
            coiled.add_interaction(action="CoiledSetup", success=False, error_message=error_message)
            print()
            print("[red]There was an error setting up Coiled to use your Azure account:\n")
            print(error_message)


def setup_with_service_principal(creds, app_name, sub_id, rg_name, rg_id, coiled_account, region):
    prompt = f"Create [green]{app_name}[/green] service principal and grant Coiled access to your Azure subscription?"
    if not Confirm.ask(prompt, default=True):
        coiled.add_interaction(action="prompt:Setup_Azure", success=False)
        return False

    enable_providers(creds, sub_id)

    print(f"  [bright_black]Creating enterprise application {app_name}...")
    app_json = az_cli_wrapper(f"az ad app create --display-name '{app_name}'")

    app_info = json.loads(app_json)
    app_id = app_info["appId"]

    print(f"  [bright_black]Resetting/retrieving credentials for {app_name} ({app_id})...")
    app_creds_json = az_cli_wrapper(f"az ad app credential reset --id {app_id}")
    app_creds = json.loads(app_creds_json)

    print(f"  [bright_black]Creating service principal for {app_name} ({app_id})...")
    sp_id = strip_output(
        az_cli_wrapper(
            f"az ad sp create --id {app_id} --query id",
            command_if_exists=f"az ad sp list --display-name '{app_name}' --query '[0].id'",
        )
    )

    print(f"  [bright_black]Creating/updating role definition {RG_ROLE_NAME}...")
    az_cli_wrapper(
        f'az role definition create --role-definition "{rg_role_def(sub_id)}"',
        command_if_exists=f'az role definition update --role-definition "{rg_role_def(sub_id)}"',
    )
    print(f"  [bright_black]Creating/updating role definition {LOG_ROLE_NAME}...")
    az_cli_wrapper(
        f'az role definition create --role-definition "{log_role_def(sub_id)}"',
        command_if_exists=f'az role definition update --role-definition "{log_role_def(sub_id)}"',
    )

    print(f"  [bright_black]Assigning '{RG_ROLE_NAME}' role to service principal on '{rg_name}' resource group...")
    az_cli_wrapper(f"az role assignment create --role '{RG_ROLE_NAME}' --scope {rg_id} --assignee {sp_id}")
    print(f"  [bright_black]Assigning '{LOG_ROLE_NAME}' role to service principal on '{rg_name}' resource group...")
    az_cli_wrapper(f"az role assignment create --role '{LOG_ROLE_NAME}' --scope {rg_id} --assignee {sp_id}")

    creds_to_submit = {
        "tenant_id": app_creds["tenant"],
        "client_id": app_creds["appId"],
        "client_secret": app_creds["password"],
    }

    print("Sending Azure credentials to Coiled... ", end="")
    submit_azure_credentials(
        coiled_account=coiled_account,
        sub_id=sub_id,
        rg_name=rg_name,
        region=region,
        creds_to_submit=creds_to_submit,
    )
    print("done!")
    print()
    print()
    print(SUCCESS_MESSAGE)

    coiled.add_interaction(action="CoiledSetup", success=True)

    return True


def submit_azure_credentials(coiled_account, sub_id, rg_name, region, creds_to_submit, check_after: bool = False):
    with coiled.Cloud(account=coiled_account) as cloud:
        setup_endpoint = f"/api/v2/cloud-credentials/{coiled_account}/azure"
        setup_data = {
            "credentials": creds_to_submit,
            "subscription_id": sub_id,
            "resource_group_name": rg_name,
            "default_region": region,
        }
        cloud._sync_request(setup_endpoint, method="POST", json=setup_data)

        if check_after:
            print("Coiled Azure credentials...")
            # did it work?
            print(cloud._sync_request(setup_endpoint))


def strip_output(output: str) -> str:
    return output.strip(' \n"')


def get_cli_path() -> Optional[str]:
    return shutil.which("az")


def az_cli_wrapper(
    command: str,
    command_if_exists: str = "",
    show_stdout: bool = False,
    interactive: bool = False,
):
    split_command = shlex.split(command)
    if split_command and split_command[0] == "az":
        del split_command[0]
    az_path = get_cli_path() or "az"
    split_command = [az_path] + split_command
    p = subprocess.run(split_command, capture_output=not interactive)

    stdout = p.stdout.decode(encoding=coiled.utils.get_encoding()) if p.stdout else None
    stderr = p.stderr.decode(encoding=coiled.utils.get_encoding(stderr=True)) if p.stderr else None

    if show_stdout:
        print(stdout)

    if not interactive and p.returncode:
        if stderr and "already" in stderr:
            if command_if_exists:
                return az_cli_wrapper(command_if_exists)

            # TODO is this always fine to ignore because it means item exists?
            return ""

        else:
            print(f"[red]az returned an error when running command `{shlex.join(split_command)}`:")
            print(Panel(escape(stderr or "")))
            setup_failure(f"az error while running {shlex.join(split_command)}, {stderr}", backend="azure")
            sys.exit(1)

    return stdout or ""


def choice_prompt(choices, prompt, exit_return=None):
    prompt_choices = [str(i) for i in range(1, len(choices) + 1)]
    for i, (_, display) in enumerate(choices):
        print(f"{i + 1}. {display}")

    exit_choice = 0
    prompt_choices.append(str(exit_choice))
    print(f"{exit_choice}. [red]Exit setup")

    print()
    choice = rich.prompt.IntPrompt.ask(prompt, choices=prompt_choices)
    if choice == exit_choice:
        return exit_return
    else:
        sub_id, _ = choices[choice - 1]
        return sub_id


def get_subscription(credentials):
    sub_client = SubscriptionClient(credentials)  # pyright: ignore[reportPossiblyUnboundVariable]

    subscriptions = [
        (subscription.display_name, subscription.subscription_id) for subscription in sub_client.subscriptions.list()
    ]

    if not subscriptions:
        print("No Azure subscriptions were found")
        return None
    elif len(subscriptions) > 1:
        print("Multiple Azure subscriptions were found.")
        choices = [(id, f"{name} (id [green]{id}[/green])") for name, id in subscriptions]
        sub_id = choice_prompt(choices, "Please select one of the available subscriptions for Coiled to use")

    else:
        sub_name, sub_id = subscriptions[0]

    return sub_id


def get_rg(credentials, sub_id: str, rg_name: str):
    resource_client = ResourceManagementClient(credentials, sub_id)  # pyright: ignore[reportPossiblyUnboundVariable]

    resource_groups = [(rg.name, rg.location, rg.id) for rg in resource_client.resource_groups.list()]

    if not resource_groups:
        print("No Azure resource groups found")
        return None, None, None
    elif len(resource_groups) > 1:
        if rg_name:
            match = [rg for rg in resource_groups if rg[0] and rg[0].lower() == rg_name.lower()]
            if match:
                return match[0]
            else:
                print(f"No resource group matches [green]{rg_name}")

        print("Multiple Azure resource groups were found.")
        choices = [((name, location, rg_id), f"{name} ({location})") for name, location, rg_id in resource_groups]
        return choice_prompt(
            choices,
            "Please select one of the available resource groups for Coiled to use",
            exit_return=(None, None, None),
        )

    else:
        return resource_groups[0]


def enable_providers(credentials, sub_id):
    resource_client = ResourceManagementClient(credentials, sub_id)  # pyright: ignore[reportPossiblyUnboundVariable]
    registered = {
        provider: resource_client.providers.get(provider).registration_state == "Registered"
        for provider in REQUIRED_PROVIDERS
    }

    if not all(registered.values()):
        print(f"Registering required providers in Azure Subscription {sub_id}...")
        print("(there's no associated cost, this just enables these services for use in the Azure subscription)")

        for provider in REQUIRED_PROVIDERS:
            if registered.get(provider):
                print(f"  [bright_black]{provider} already registered")
            else:
                print(f"  Enabling [green]{provider}[/green]... ", end="")
                resource_client.providers.register(provider)
                print("successfully registered!")
        print()


def get_temp_delegation_display(
    azure_user, coiled_account, sub_id, rg_name, expiration, last_refresh, active: bool = True
):
    now = datetime.datetime.now(tz=datetime.timezone.utc)
    remaining_time = expiration - now
    refresh_since = now - last_refresh

    expiration_message = (
        f"in [bold]{remaining_time.seconds // 60} minutes[/bold]" if active else f"at [bold]{expiration:%c %Z}[/bold]"
    )
    last_refresh_message = (
        f"{refresh_since.seconds // 60}m{refresh_since.seconds % 60}s ago at {last_refresh:%c %Z}"
        if active
        else f"{last_refresh:%c %Z}"
    )

    status = "[green]active[/green]" if active else "[red]inactive[/red]"

    return Panel(
        f"""Access delegation is [bold]{status}[/bold] for Azure!

Coiled is now able to temporarily act as you to create and manage resources in your resource group.

Access will expire {expiration_message} if not refreshed.
  Last refresh:  {last_refresh_message}

Azure user principal:  [bold]{azure_user["userPrincipalName"]}[/bold]
                       (id {azure_user["id"]})
Azure resource group:  [bold]{rg_name}[/bold]
                       (subscription {sub_id})
Coiled workspace:      [bold]{coiled_account}[/bold]

Use Control-C to stop refreshing access delegation""",
        width=80,
    )


def get_temporary_creds_to_submit(local_credentials):
    scope = "https://management.azure.com/.default"
    token_creds = local_credentials.get_token(scope)

    creds_to_submit = {
        "token_scope": scope,
        "token_value": token_creds.token,
        "token_expiration": token_creds.expires_on,
    }
    expiration = datetime.datetime.fromtimestamp(token_creds.expires_on, tz=datetime.timezone.utc)

    return creds_to_submit, expiration


def ship_token_creds(local_credentials, coiled_account, sub_id, rg_name, region):
    azure_user = whoami(local_credentials)

    creds_to_submit, expiration = get_temporary_creds_to_submit(local_credentials)
    last_refresh = datetime.datetime.now(tz=datetime.timezone.utc)
    submit_azure_credentials(
        coiled_account=coiled_account,
        sub_id=sub_id,
        rg_name=rg_name,
        region=region,
        creds_to_submit=creds_to_submit,
        check_after=False,
    )
    display = get_temp_delegation_display(
        azure_user=azure_user,
        coiled_account=coiled_account,
        sub_id=sub_id,
        rg_name=rg_name,
        expiration=expiration,
        last_refresh=last_refresh,
    )

    with Live(display) as live:
        try:
            while True:
                # update temporary credentials if they'll expire within 7 minutes
                soon = datetime.datetime.now(tz=datetime.timezone.utc) + datetime.timedelta(minutes=7)
                if expiration < soon:
                    creds_to_submit, expiration = get_temporary_creds_to_submit(local_credentials)
                    last_refresh = datetime.datetime.now(tz=datetime.timezone.utc)
                    submit_azure_credentials(
                        coiled_account=coiled_account,
                        sub_id=sub_id,
                        rg_name=rg_name,
                        region=region,
                        creds_to_submit=creds_to_submit,
                        check_after=False,
                    )

                display = get_temp_delegation_display(
                    azure_user=azure_user,
                    coiled_account=coiled_account,
                    sub_id=sub_id,
                    rg_name=rg_name,
                    expiration=expiration,
                    last_refresh=last_refresh,
                )
                live.update(display)
                time.sleep(0.5)

        except KeyboardInterrupt:
            display = get_temp_delegation_display(
                azure_user=azure_user,
                coiled_account=coiled_account,
                sub_id=sub_id,
                rg_name=rg_name,
                expiration=expiration,
                last_refresh=last_refresh,
                active=False,
            )
            live.update(display)
            live.stop()
            print(
                "Temporary access to your Azure resources will no longer be refreshed.\n"
                f"Unless you restart the process to refresh access, any running clusters will be automatically stopped "
                f"5 minutes before the current access token expires at {expiration:%c %Z}."
            )
            return


def whoami(local_credentials):
    try:
        # note that this prints error if there aren't creds
        token_creds = local_credentials.get_token("https://graph.microsoft.com/.default")
    except ClientAuthenticationError:
        return None
    with httpx.Client(http2=True) as client:
        result = client.get(
            "https://graph.microsoft.com/v1.0/me?$select=id,userPrincipalName",
            headers={"Authorization": f"Bearer {token_creds.token}"},
        )
    return result.json()
